package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.ConstraintFormula;
import org.eclipse.jdt.internal.compiler.lookup.ConstraintTypeFormula;
import org.eclipse.jdt.internal.compiler.lookup.InferenceContext18;
import org.eclipse.jdt.internal.compiler.lookup.InferenceFailureException;
import org.eclipse.jdt.internal.compiler.lookup.InferenceSubstitution;
import org.eclipse.jdt.internal.compiler.lookup.InferenceVariable;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReductionResult;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Substitution;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBound;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.WildcardBinding;

class BoundSet {
  static final org.eclipse.jdt.internal.compiler.lookup.BoundSet TRUE = new org.eclipse.jdt.internal.compiler.lookup.BoundSet();
  
  static final org.eclipse.jdt.internal.compiler.lookup.BoundSet FALSE = new org.eclipse.jdt.internal.compiler.lookup.BoundSet();
  
  HashMap<InferenceVariable, ThreeSets> boundsPerVariable = new HashMap<>();
  
  HashMap<ParameterizedTypeBinding, ParameterizedTypeBinding> captures = new HashMap<>();
  
  Set<InferenceVariable> inThrows = new HashSet<>();
  
  private TypeBound[] incorporatedBounds = new TypeBound[0];
  
  private TypeBound[] unincorporatedBounds = new TypeBound[1024];
  
  private int unincorporatedBoundsCount = 0;
  
  private TypeBound[] mostRecentBounds = new TypeBound[4];
  
  public void addBoundsFromTypeParameters(InferenceContext18 context, TypeVariableBinding[] typeParameters, InferenceVariable[] variables) {
    int length = typeParameters.length;
    for (int i = 0; i < length; i++) {
      TypeVariableBinding typeParameter = typeParameters[i];
      InferenceVariable variable = variables[i];
      TypeBound[] someBounds = typeParameter.getTypeBounds(variable, new InferenceSubstitution(context));
      boolean hasProperBound = false;
      if (someBounds.length > 0)
        hasProperBound = addBounds(someBounds, context.environment); 
      if (!hasProperBound)
        addBound(new TypeBound(variable, (TypeBinding)context.object, 2), context.environment); 
    } 
  }
  
  public TypeBound[] flatten() {
    int size = 0;
    Iterator<ThreeSets> outerIt = this.boundsPerVariable.values().iterator();
    while (outerIt.hasNext())
      size += ((ThreeSets)outerIt.next()).size(); 
    TypeBound[] collected = new TypeBound[size];
    if (size == 0)
      return collected; 
    outerIt = this.boundsPerVariable.values().iterator();
    int idx = 0;
    while (outerIt.hasNext())
      idx = ((ThreeSets)outerIt.next()).flattenInto(collected, idx); 
    return collected;
  }
  
  public org.eclipse.jdt.internal.compiler.lookup.BoundSet copy() {
    org.eclipse.jdt.internal.compiler.lookup.BoundSet copy = new org.eclipse.jdt.internal.compiler.lookup.BoundSet();
    Iterator<Map.Entry<InferenceVariable, ThreeSets>> setsIterator = this.boundsPerVariable.entrySet().iterator();
    while (setsIterator.hasNext()) {
      Map.Entry<InferenceVariable, ThreeSets> entry = setsIterator.next();
      copy.boundsPerVariable.put(entry.getKey(), ((ThreeSets)entry.getValue()).copy());
    } 
    copy.inThrows.addAll(this.inThrows);
    copy.captures.putAll(this.captures);
    System.arraycopy(this.incorporatedBounds, 0, copy.incorporatedBounds = new TypeBound[this.incorporatedBounds.length], 0, this.incorporatedBounds.length);
    System.arraycopy(this.unincorporatedBounds, 0, copy.unincorporatedBounds = new TypeBound[this.unincorporatedBounds.length], 0, this.unincorporatedBounds.length);
    copy.unincorporatedBoundsCount = this.unincorporatedBoundsCount;
    return copy;
  }
  
  public void addBound(TypeBound bound, LookupEnvironment environment) {
    if (bound.relation == 2 && bound.right.id == 1)
      return; 
    if (bound.left == bound.right)
      return; 
    for (int recent = 0; recent < 4; recent++) {
      if (bound.equals(this.mostRecentBounds[recent])) {
        if (environment.globalOptions.isAnnotationBasedNullAnalysisEnabled) {
          TypeBound existing = this.mostRecentBounds[recent];
          long boundNullBits = bound.right.tagBits & 0x180000000000000L;
          long existingNullBits = existing.right.tagBits & 0x180000000000000L;
          if (boundNullBits != existingNullBits)
            if (existingNullBits == 0L) {
              existing.right = bound.right;
            } else if (boundNullBits != 0L) {
              existing.right = environment.createAnnotatedType(existing.right, environment.nullAnnotationsFromTagBits(boundNullBits));
            }  
        } 
        return;
      } 
    } 
    this.mostRecentBounds[3] = this.mostRecentBounds[2];
    this.mostRecentBounds[2] = this.mostRecentBounds[1];
    this.mostRecentBounds[1] = this.mostRecentBounds[0];
    this.mostRecentBounds[0] = bound;
    InferenceVariable variable = bound.left.prototype();
    ThreeSets three = this.boundsPerVariable.get(variable);
    if (three == null)
      this.boundsPerVariable.put(variable, three = new ThreeSets(this)); 
    if (three.addBound(bound)) {
      int unincorporatedBoundsLength = this.unincorporatedBounds.length;
      if (this.unincorporatedBoundsCount >= unincorporatedBoundsLength)
        System.arraycopy(this.unincorporatedBounds, 0, this.unincorporatedBounds = new TypeBound[unincorporatedBoundsLength * 2], 0, unincorporatedBoundsLength); 
      this.unincorporatedBounds[this.unincorporatedBoundsCount++] = bound;
      TypeBinding typeBinding = bound.right;
      if (bound.relation == 4 && typeBinding.isProperType(true))
        three.setInstantiation(typeBinding, variable, environment); 
      if (bound.right instanceof InferenceVariable) {
        InferenceVariable rightIV = (InferenceVariable)bound.right.prototype();
        three = this.boundsPerVariable.get(rightIV);
        if (three == null)
          this.boundsPerVariable.put(rightIV, three = new ThreeSets(this)); 
        if (three.inverseBounds == null)
          three.inverseBounds = new HashMap<>(); 
        three.inverseBounds.put(rightIV, bound);
      } 
    } 
  }
  
  private boolean addBounds(TypeBound[] newBounds, LookupEnvironment environment) {
    boolean hasProperBound = false;
    for (int i = 0; i < newBounds.length; i++) {
      addBound(newBounds[i], environment);
      hasProperBound |= newBounds[i].isBound();
    } 
    return hasProperBound;
  }
  
  public void addBounds(org.eclipse.jdt.internal.compiler.lookup.BoundSet that, LookupEnvironment environment) {
    if (that == null || environment == null)
      return; 
    addBounds(that.flatten(), environment);
  }
  
  public boolean isInstantiated(InferenceVariable inferenceVariable) {
    ThreeSets three = this.boundsPerVariable.get(inferenceVariable.prototype());
    if (three != null)
      return (three.instantiation != null); 
    return false;
  }
  
  public TypeBinding getInstantiation(InferenceVariable inferenceVariable, LookupEnvironment environment) {
    ThreeSets three = this.boundsPerVariable.get(inferenceVariable.prototype());
    if (three != null) {
      TypeBinding instantiation = three.instantiation;
      if (environment != null && environment.globalOptions.isAnnotationBasedNullAnalysisEnabled && 
        instantiation != null && (instantiation.tagBits & 0x180000000000000L) == 0L)
        return three.combineAndUseNullHints(instantiation, inferenceVariable.nullHints, environment); 
      return instantiation;
    } 
    return null;
  }
  
  public int numUninstantiatedVariables(InferenceVariable[] variables) {
    int num = 0;
    for (int i = 0; i < variables.length; i++) {
      if (!isInstantiated(variables[i]))
        num++; 
    } 
    return num;
  }
  
  boolean incorporate(InferenceContext18 context) throws InferenceFailureException {
    if (this.unincorporatedBoundsCount == 0 && this.captures.size() == 0)
      return true; 
    do {
      TypeBound[] freshBounds;
      System.arraycopy(this.unincorporatedBounds, 0, freshBounds = new TypeBound[this.unincorporatedBoundsCount], 0, this.unincorporatedBoundsCount);
      this.unincorporatedBoundsCount = 0;
      if (!incorporate(context, this.incorporatedBounds, freshBounds))
        return false; 
      if (!incorporate(context, freshBounds, freshBounds))
        return false; 
      int incorporatedLength = this.incorporatedBounds.length;
      int unincorporatedLength = freshBounds.length;
      TypeBound[] aggregate = new TypeBound[incorporatedLength + unincorporatedLength];
      System.arraycopy(this.incorporatedBounds, 0, aggregate, 0, incorporatedLength);
      System.arraycopy(freshBounds, 0, aggregate, incorporatedLength, unincorporatedLength);
      this.incorporatedBounds = aggregate;
    } while (this.unincorporatedBoundsCount > 0);
    return true;
  }
  
  boolean incorporate(InferenceContext18 context, TypeBound[] first, TypeBound[] next) throws InferenceFailureException {
    boolean analyzeNull = context.environment.globalOptions.isAnnotationBasedNullAnalysisEnabled;
    ConstraintTypeFormula[] mostRecentFormulas = new ConstraintTypeFormula[4];
    for (int i = 0, iLength = first.length; i < iLength; i++) {
      TypeBound boundI = first[i];
      for (int j = 0, jLength = next.length; j < jLength; j++) {
        TypeBound boundJ = next[j];
        if (boundI != boundJ) {
          int iteration = 1;
          do {
            ConstraintTypeFormula newConstraint = null;
            boolean deriveTypeArgumentConstraints = false;
            if (iteration == 2) {
              TypeBound typeBound = boundI;
              boundI = boundJ;
              boundJ = typeBound;
            } 
            switch (boundI.relation) {
              case 4:
                switch (boundJ.relation) {
                  case 4:
                    newConstraint = combineSameSame(boundI, boundJ);
                    break;
                  case 2:
                  case 3:
                    newConstraint = combineSameSubSuper(boundI, boundJ);
                    break;
                } 
                break;
              case 2:
                switch (boundJ.relation) {
                  case 4:
                    newConstraint = combineSameSubSuper(boundJ, boundI);
                    break;
                  case 3:
                    newConstraint = combineSuperAndSub(boundJ, boundI);
                    break;
                  case 2:
                    newConstraint = combineEqualSupers(boundI, boundJ);
                    deriveTypeArgumentConstraints = TypeBinding.equalsEquals((TypeBinding)boundI.left, (TypeBinding)boundJ.left);
                    break;
                } 
                break;
              case 3:
                switch (boundJ.relation) {
                  case 4:
                    newConstraint = combineSameSubSuper(boundJ, boundI);
                    break;
                  case 2:
                    newConstraint = combineSuperAndSub(boundI, boundJ);
                    break;
                  case 3:
                    newConstraint = combineEqualSupers(boundI, boundJ);
                    break;
                } 
                break;
            } 
            if (newConstraint != null)
              if (newConstraint.left == newConstraint.right) {
                newConstraint = null;
              } else if (newConstraint.equalsEquals(mostRecentFormulas[0]) || newConstraint.equalsEquals(mostRecentFormulas[1]) || 
                newConstraint.equalsEquals(mostRecentFormulas[2]) || newConstraint.equalsEquals(mostRecentFormulas[3])) {
                newConstraint = null;
              }  
            if (newConstraint != null) {
              mostRecentFormulas[3] = mostRecentFormulas[2];
              mostRecentFormulas[2] = mostRecentFormulas[1];
              mostRecentFormulas[1] = mostRecentFormulas[0];
              mostRecentFormulas[0] = newConstraint;
              if (!reduceOneConstraint(context, (ConstraintFormula)newConstraint))
                return false; 
              if (analyzeNull) {
                long nullHints = (newConstraint.left.tagBits | newConstraint.right.tagBits) & 0x180000000000000L;
                if (nullHints != 0L && (
                  TypeBinding.equalsEquals((TypeBinding)boundI.left, (TypeBinding)boundJ.left) || (
                  boundI.relation == 4 && TypeBinding.equalsEquals(boundI.right, (TypeBinding)boundJ.left)) || (
                  boundJ.relation == 4 && TypeBinding.equalsEquals((TypeBinding)boundI.left, boundJ.right)))) {
                  boundI.nullHints |= nullHints;
                  boundJ.nullHints |= nullHints;
                } 
              } 
            } 
            ConstraintTypeFormula[] arrayOfConstraintTypeFormula = deriveTypeArgumentConstraints ? deriveTypeArgumentConstraints(boundI, boundJ) : null;
            if (arrayOfConstraintTypeFormula != null)
              for (int k = 0, length = arrayOfConstraintTypeFormula.length; k < length; k++) {
                if (!reduceOneConstraint(context, (ConstraintFormula)arrayOfConstraintTypeFormula[k]))
                  return false; 
              }  
            if (iteration != 2)
              continue; 
            TypeBound boundX = boundI;
            boundI = boundJ;
            boundJ = boundX;
          } while (first != next && ++iteration <= 2);
        } 
      } 
    } 
    Iterator<Map.Entry<ParameterizedTypeBinding, ParameterizedTypeBinding>> captIter = this.captures.entrySet().iterator();
    while (captIter.hasNext()) {
      Map.Entry<ParameterizedTypeBinding, ParameterizedTypeBinding> capt = captIter.next();
      ParameterizedTypeBinding gAlpha = capt.getKey();
      ParameterizedTypeBinding gA = capt.getValue();
      ReferenceBinding g = (ReferenceBinding)gA.original();
      TypeVariableBinding[] parameters = g.typeVariables();
      InferenceVariable[] alphas = new InferenceVariable[gAlpha.arguments.length];
      System.arraycopy(gAlpha.arguments, 0, alphas, 0, alphas.length);
      Object object = new Object(this, context.environment, alphas, context.currentInvocation, parameters);
      for (int j = 0, length = parameters.length; j < length; j++) {
        TypeVariableBinding pi = parameters[j];
        InferenceVariable alpha = (InferenceVariable)gAlpha.arguments[j];
        addBounds(pi.getTypeBounds(alpha, (InferenceSubstitution)object), context.environment);
        TypeBinding ai = gA.arguments[j];
        if (ai instanceof WildcardBinding) {
          WildcardBinding wildcardBinding = (WildcardBinding)ai;
          TypeBinding t = wildcardBinding.bound;
          ThreeSets three = this.boundsPerVariable.get(alpha.prototype());
          if (three != null) {
            if (three.sameBounds != null) {
              Iterator<TypeBound> it = three.sameBounds.iterator();
              while (it.hasNext()) {
                TypeBound bound = it.next();
                if (bound.right instanceof org.eclipse.jdt.internal.compiler.lookup.CaptureBinding && bound.right.isProperType(true))
                  continue; 
                if (!(bound.right instanceof InferenceVariable))
                  return false; 
              } 
            } 
            if (three.subBounds != null && pi.firstBound != null) {
              Iterator<TypeBound> it = three.subBounds.iterator();
              while (it.hasNext()) {
                TypeBound bound = it.next();
                if (!(bound.right instanceof InferenceVariable)) {
                  TypeBinding bi, r = bound.right;
                  TypeBinding bi1 = pi.firstBound;
                  ReferenceBinding[] otherBounds = pi.superInterfaces;
                  if (otherBounds == Binding.NO_SUPERINTERFACES) {
                    bi = bi1;
                  } else {
                    int n = otherBounds.length + 1;
                    ReferenceBinding[] allBounds = new ReferenceBinding[n];
                    allBounds[0] = (ReferenceBinding)bi1;
                    System.arraycopy(otherBounds, 0, allBounds, 1, n - 1);
                    bi = context.environment.createIntersectionType18(allBounds);
                  } 
                  addTypeBoundsFromWildcardBound(context, (InferenceSubstitution)object, wildcardBinding.boundKind, t, r, bi);
                } 
              } 
            } 
            if (three.superBounds != null) {
              Iterator<TypeBound> it = three.superBounds.iterator();
              while (it.hasNext()) {
                TypeBound bound = it.next();
                if (!(bound.right instanceof InferenceVariable)) {
                  if (wildcardBinding.boundKind == 2) {
                    reduceOneConstraint(context, (ConstraintFormula)ConstraintTypeFormula.create(bound.right, t, 2));
                    continue;
                  } 
                  return false;
                } 
              } 
            } 
          } 
        } else {
          addBound(new TypeBound(alpha, ai, 4), context.environment);
        } 
      } 
    } 
    this.captures.clear();
    return true;
  }
  
  void addTypeBoundsFromWildcardBound(InferenceContext18 context, InferenceSubstitution theta, int boundKind, TypeBinding t, TypeBinding r, TypeBinding bi) throws InferenceFailureException {
    ConstraintTypeFormula constraintTypeFormula;
    ConstraintFormula formula = null;
    if (boundKind == 1) {
      if (bi.id == 1)
        constraintTypeFormula = ConstraintTypeFormula.create(t, r, 2); 
      if (t.id == 1)
        constraintTypeFormula = ConstraintTypeFormula.create(theta.substitute((Substitution)theta, bi), r, 2); 
    } else {
      constraintTypeFormula = ConstraintTypeFormula.create(theta.substitute((Substitution)theta, bi), r, 2);
    } 
    if (constraintTypeFormula != null)
      reduceOneConstraint(context, (ConstraintFormula)constraintTypeFormula); 
  }
  
  private ConstraintTypeFormula combineSameSame(TypeBound boundS, TypeBound boundT) {
    if (TypeBinding.equalsEquals((TypeBinding)boundS.left, (TypeBinding)boundT.left))
      return ConstraintTypeFormula.create(boundS.right, boundT.right, 4, !(!boundS.isSoft && !boundT.isSoft)); 
    ConstraintTypeFormula newConstraint = combineSameSameWithProperType(boundS, boundT);
    if (newConstraint != null)
      return newConstraint; 
    newConstraint = combineSameSameWithProperType(boundT, boundS);
    if (newConstraint != null)
      return newConstraint; 
    return null;
  }
  
  private ConstraintTypeFormula combineSameSameWithProperType(TypeBound boundLeft, TypeBound boundRight) {
    TypeBinding u = boundLeft.right;
    if (u.isProperType(true)) {
      InferenceVariable alpha = boundLeft.left;
      InferenceVariable inferenceVariable1 = boundRight.left;
      TypeBinding right = boundRight.right.substituteInferenceVariable(alpha, u);
      return ConstraintTypeFormula.create((TypeBinding)inferenceVariable1, right, 4, !(!boundLeft.isSoft && !boundRight.isSoft));
    } 
    return null;
  }
  
  private ConstraintTypeFormula combineSameSubSuper(TypeBound boundS, TypeBound boundT) {
    InferenceVariable alpha = boundS.left;
    TypeBinding s = boundS.right;
    if (TypeBinding.equalsEquals((TypeBinding)alpha, (TypeBinding)boundT.left)) {
      TypeBinding t = boundT.right;
      return ConstraintTypeFormula.create(s, t, boundT.relation, !(!boundT.isSoft && !boundS.isSoft));
    } 
    if (TypeBinding.equalsEquals((TypeBinding)alpha, boundT.right)) {
      InferenceVariable inferenceVariable = boundT.left;
      return ConstraintTypeFormula.create((TypeBinding)inferenceVariable, s, boundT.relation, !(!boundT.isSoft && !boundS.isSoft));
    } 
    if (boundS.right instanceof InferenceVariable) {
      alpha = (InferenceVariable)boundS.right;
      InferenceVariable inferenceVariable = boundS.left;
      if (TypeBinding.equalsEquals((TypeBinding)alpha, (TypeBinding)boundT.left)) {
        TypeBinding t = boundT.right;
        return ConstraintTypeFormula.create((TypeBinding)inferenceVariable, t, boundT.relation, !(!boundT.isSoft && !boundS.isSoft));
      } 
      if (TypeBinding.equalsEquals((TypeBinding)alpha, boundT.right)) {
        InferenceVariable inferenceVariable1 = boundT.left;
        return ConstraintTypeFormula.create((TypeBinding)inferenceVariable1, (TypeBinding)inferenceVariable, boundT.relation, !(!boundT.isSoft && !boundS.isSoft));
      } 
    } 
    TypeBinding u = boundS.right;
    if (u.isProperType(true)) {
      boolean substitute = TypeBinding.equalsEquals((TypeBinding)alpha, (TypeBinding)boundT.left);
      TypeBinding left = substitute ? u : (TypeBinding)boundT.left;
      TypeBinding right = boundT.right.substituteInferenceVariable(alpha, u);
      substitute |= TypeBinding.notEquals(right, boundT.right);
      if (substitute)
        return ConstraintTypeFormula.create(left, right, boundT.relation, !(!boundT.isSoft && !boundS.isSoft)); 
    } 
    return null;
  }
  
  private ConstraintTypeFormula combineSuperAndSub(TypeBound boundS, TypeBound boundT) {
    InferenceVariable alpha = boundS.left;
    if (TypeBinding.equalsEquals((TypeBinding)alpha, (TypeBinding)boundT.left))
      return ConstraintTypeFormula.create(boundS.right, boundT.right, 2, !(!boundT.isSoft && !boundS.isSoft)); 
    if (boundS.right instanceof InferenceVariable) {
      alpha = (InferenceVariable)boundS.right;
      if (TypeBinding.equalsEquals((TypeBinding)alpha, boundT.right))
        return ConstraintTypeFormula.create((TypeBinding)boundS.left, (TypeBinding)boundT.left, 3, !(!boundT.isSoft && !boundS.isSoft)); 
    } 
    return null;
  }
  
  private ConstraintTypeFormula combineEqualSupers(TypeBound boundS, TypeBound boundT) {
    if (TypeBinding.equalsEquals((TypeBinding)boundS.left, boundT.right))
      return ConstraintTypeFormula.create((TypeBinding)boundT.left, boundS.right, boundS.relation, !(!boundT.isSoft && !boundS.isSoft)); 
    if (TypeBinding.equalsEquals(boundS.right, (TypeBinding)boundT.left))
      return ConstraintTypeFormula.create((TypeBinding)boundS.left, boundT.right, boundS.relation, !(!boundT.isSoft && !boundS.isSoft)); 
    return null;
  }
  
  private ConstraintTypeFormula[] deriveTypeArgumentConstraints(TypeBound boundS, TypeBound boundT) {
    TypeBinding[] supers = superTypesWithCommonGenericType(boundS.right, boundT.right);
    if (supers != null)
      return typeArgumentEqualityConstraints(supers[0], supers[1], !(!boundS.isSoft && !boundT.isSoft)); 
    return null;
  }
  
  private ConstraintTypeFormula[] typeArgumentEqualityConstraints(TypeBinding s, TypeBinding t, boolean isSoft) {
    if (s == null || s.kind() != 260 || t == null || t.kind() != 260)
      return null; 
    if (TypeBinding.equalsEquals(s, t))
      return null; 
    TypeBinding[] sis = s.typeArguments();
    TypeBinding[] tis = t.typeArguments();
    if (sis == null || tis == null || sis.length != tis.length)
      return null; 
    List<ConstraintTypeFormula> result = new ArrayList<>();
    for (int i = 0; i < sis.length; i++) {
      TypeBinding si = sis[i];
      TypeBinding ti = tis[i];
      if (!si.isWildcard() && !ti.isWildcard() && !TypeBinding.equalsEquals(si, ti))
        result.add(ConstraintTypeFormula.create(si, ti, 4, isSoft)); 
    } 
    if (result.size() > 0)
      return result.<ConstraintTypeFormula>toArray(new ConstraintTypeFormula[result.size()]); 
    return null;
  }
  
  public boolean reduceOneConstraint(InferenceContext18 context, ConstraintFormula currentConstraint) throws InferenceFailureException {
    Object result = currentConstraint.reduce(context);
    if (result == ReductionResult.FALSE)
      return false; 
    if (result == ReductionResult.TRUE)
      return true; 
    if (result == currentConstraint)
      throw new IllegalStateException("Failed to reduce constraint formula"); 
    if (result != null)
      if (result instanceof ConstraintFormula) {
        if (!reduceOneConstraint(context, (ConstraintFormula)result))
          return false; 
      } else if (result instanceof ConstraintFormula[]) {
        ConstraintFormula[] resultArray = (ConstraintFormula[])result;
        for (int i = 0; i < resultArray.length; i++) {
          if (!reduceOneConstraint(context, resultArray[i]))
            return false; 
        } 
      } else {
        addBound((TypeBound)result, context.environment);
      }  
    return true;
  }
  
  public boolean dependsOnResolutionOf(InferenceVariable alpha, InferenceVariable beta) {
    alpha = alpha.prototype();
    beta = beta.prototype();
    if (TypeBinding.equalsEquals((TypeBinding)alpha, (TypeBinding)beta))
      return true; 
    Iterator<Map.Entry<ParameterizedTypeBinding, ParameterizedTypeBinding>> captureIter = this.captures.entrySet().iterator();
    boolean betaIsInCaptureLhs = false;
    while (captureIter.hasNext()) {
      Map.Entry<ParameterizedTypeBinding, ParameterizedTypeBinding> entry = captureIter.next();
      ParameterizedTypeBinding g = entry.getKey();
      for (int i = 0; i < g.arguments.length; i++) {
        if (TypeBinding.equalsEquals(g.arguments[i], (TypeBinding)alpha)) {
          ParameterizedTypeBinding captured = entry.getValue();
          if (captured.mentionsAny(new TypeBinding[] { (TypeBinding)beta }, -1))
            return true; 
          if (g.mentionsAny(new TypeBinding[] { (TypeBinding)beta }, i))
            return true; 
        } else if (TypeBinding.equalsEquals(g.arguments[i], (TypeBinding)beta)) {
          betaIsInCaptureLhs = true;
        } 
      } 
    } 
    if (betaIsInCaptureLhs) {
      ThreeSets sets = this.boundsPerVariable.get(beta);
      if (sets != null && sets.hasDependency(alpha))
        return true; 
    } else {
      ThreeSets sets = this.boundsPerVariable.get(alpha);
      if (sets != null && sets.hasDependency(beta))
        return true; 
    } 
    return false;
  }
  
  List<Set<InferenceVariable>> computeConnectedComponents(InferenceVariable[] inferenceVariables) {
    Map<InferenceVariable, Set<InferenceVariable>> allEdges = new HashMap<>();
    for (int i = 0; i < inferenceVariables.length; i++) {
      InferenceVariable iv1 = inferenceVariables[i];
      HashSet<InferenceVariable> targetSet = new HashSet<>();
      allEdges.put(iv1, targetSet);
      for (int k = 0; k < i; k++) {
        InferenceVariable iv2 = inferenceVariables[k];
        if (dependsOnResolutionOf(iv1, iv2) || dependsOnResolutionOf(iv2, iv1)) {
          targetSet.add(iv2);
          ((Set<InferenceVariable>)allEdges.get(iv2)).add(iv1);
        } 
      } 
    } 
    Set<InferenceVariable> visited = new HashSet<>();
    List<Set<InferenceVariable>> allComponents = new ArrayList<>();
    byte b;
    int j;
    InferenceVariable[] arrayOfInferenceVariable;
    for (j = (arrayOfInferenceVariable = inferenceVariables).length, b = 0; b < j; ) {
      InferenceVariable inferenceVariable = arrayOfInferenceVariable[b];
      Set<InferenceVariable> component = new HashSet<>();
      addConnected(component, inferenceVariable, allEdges, visited);
      if (!component.isEmpty())
        allComponents.add(component); 
      b++;
    } 
    return allComponents;
  }
  
  private void addConnected(Set<InferenceVariable> component, InferenceVariable seed, Map<InferenceVariable, Set<InferenceVariable>> allEdges, Set<InferenceVariable> visited) {
    if (visited.add(seed)) {
      component.add(seed);
      for (InferenceVariable next : allEdges.get(seed))
        addConnected(component, next, allEdges, visited); 
    } 
  }
  
  public boolean hasCaptureBound(Set<InferenceVariable> variableSet) {
    Iterator<ParameterizedTypeBinding> captureIter = this.captures.keySet().iterator();
    while (captureIter.hasNext()) {
      ParameterizedTypeBinding g = captureIter.next();
      for (int i = 0; i < g.arguments.length; i++) {
        if (variableSet.contains(g.arguments[i]))
          return true; 
      } 
    } 
    return false;
  }
  
  public boolean hasOnlyTrivialExceptionBounds(InferenceVariable variable, TypeBinding[] upperBounds) {
    if (upperBounds != null)
      for (int i = 0; i < upperBounds.length; i++) {
        switch ((upperBounds[i]).id) {
          case 1:
          case 21:
          case 25:
            break;
          default:
            return false;
        } 
      }  
    return true;
  }
  
  public TypeBinding[] upperBounds(InferenceVariable variable, boolean onlyProper) {
    ThreeSets three = this.boundsPerVariable.get(variable.prototype());
    if (three == null || three.subBounds == null)
      return Binding.NO_TYPES; 
    return three.upperBounds(onlyProper, variable);
  }
  
  TypeBinding[] lowerBounds(InferenceVariable variable, boolean onlyProper) {
    ThreeSets three = this.boundsPerVariable.get(variable.prototype());
    if (three == null || three.superBounds == null)
      return Binding.NO_TYPES; 
    return three.lowerBounds(onlyProper, variable);
  }
  
  public String toString() {
    StringBuffer buf = new StringBuffer("Type Bounds:\n");
    TypeBound[] flattened = flatten();
    for (int i = 0; i < flattened.length; i++)
      buf.append('\t').append(flattened[i].toString()).append('\n'); 
    buf.append("Capture Bounds:\n");
    Iterator<Map.Entry<ParameterizedTypeBinding, ParameterizedTypeBinding>> captIter = this.captures.entrySet().iterator();
    while (captIter.hasNext()) {
      Map.Entry<ParameterizedTypeBinding, ParameterizedTypeBinding> capt = captIter.next();
      String lhs = String.valueOf(((TypeBinding)capt.getKey()).shortReadableName());
      String rhs = String.valueOf(((TypeBinding)capt.getValue()).shortReadableName());
      buf.append('\t').append(lhs).append(" = capt(").append(rhs).append(")\n");
    } 
    return buf.toString();
  }
  
  public TypeBinding findWrapperTypeBound(InferenceVariable variable) {
    ThreeSets three = this.boundsPerVariable.get(variable.prototype());
    if (three == null)
      return null; 
    return three.findSingleWrapperType();
  }
  
  public boolean condition18_5_2_bullet_3_3_1(InferenceVariable alpha, TypeBinding targetType) {
    if (targetType.isBaseType())
      return false; 
    if (InferenceContext18.parameterizedWithWildcard(targetType) != null)
      return false; 
    ThreeSets ts = this.boundsPerVariable.get(alpha.prototype());
    if (ts == null)
      return false; 
    if (ts.sameBounds != null) {
      Iterator<TypeBound> bounds = ts.sameBounds.iterator();
      while (bounds.hasNext()) {
        TypeBound bound = bounds.next();
        if (InferenceContext18.parameterizedWithWildcard(bound.right) != null)
          return true; 
      } 
    } 
    if (ts.superBounds != null) {
      Iterator<TypeBound> bounds = ts.superBounds.iterator();
      while (bounds.hasNext()) {
        TypeBound bound = bounds.next();
        if (InferenceContext18.parameterizedWithWildcard(bound.right) != null)
          return true; 
      } 
    } 
    if (ts.superBounds != null) {
      ArrayList<TypeBound> superBounds = new ArrayList<>(ts.superBounds);
      int len = superBounds.size();
      for (int i = 0; i < len; i++) {
        TypeBinding s1 = ((TypeBound)superBounds.get(i)).right;
        for (int j = i + 1; j < len; j++) {
          TypeBinding s2 = ((TypeBound)superBounds.get(j)).right;
          TypeBinding[] supers = superTypesWithCommonGenericType(s1, s2);
          if (supers != null)
            if (supers[0].isProperType(true) && supers[1].isProperType(true) && !TypeBinding.equalsEquals(supers[0], supers[1]))
              return true;  
        } 
      } 
    } 
    return false;
  }
  
  public boolean condition18_5_2_bullet_3_3_2(InferenceVariable alpha, TypeBinding targetType, InferenceContext18 ctx18) {
    if (!targetType.isParameterizedType())
      return false; 
    TypeBinding g = targetType.original();
    ThreeSets ts = this.boundsPerVariable.get(alpha.prototype());
    if (ts == null)
      return false; 
    if (ts.sameBounds != null) {
      Iterator<TypeBound> boundIterator = ts.sameBounds.iterator();
      while (boundIterator.hasNext()) {
        TypeBound b = boundIterator.next();
        if (superOnlyRaw(g, b.right, ctx18.environment))
          return true; 
      } 
    } 
    if (ts.superBounds != null) {
      Iterator<TypeBound> boundIterator = ts.superBounds.iterator();
      while (boundIterator.hasNext()) {
        TypeBound b = boundIterator.next();
        if (superOnlyRaw(g, b.right, ctx18.environment))
          return true; 
      } 
    } 
    return false;
  }
  
  private boolean superOnlyRaw(TypeBinding g, TypeBinding s, LookupEnvironment env) {
    if (s instanceof InferenceVariable)
      return false; 
    TypeBinding superType = s.findSuperTypeOriginatingFrom(g);
    if (superType != null && !superType.isParameterizedType())
      return s.isCompatibleWith(env.convertToRawType(g, false)); 
    return false;
  }
  
  protected TypeBinding[] superTypesWithCommonGenericType(TypeBinding s, TypeBinding t) {
    if (s == null || s.id == 1 || t == null || t.id == 1)
      return null; 
    if (TypeBinding.equalsEquals(s.original(), t.original()))
      return new TypeBinding[] { s, t }; 
    TypeBinding tSuper = t.findSuperTypeOriginatingFrom(s);
    if (tSuper != null)
      return new TypeBinding[] { s, tSuper }; 
    TypeBinding[] result = superTypesWithCommonGenericType((TypeBinding)s.superclass(), t);
    if (result != null)
      return result; 
    ReferenceBinding[] superInterfaces = s.superInterfaces();
    if (superInterfaces != null)
      for (int i = 0; i < superInterfaces.length; i++) {
        result = superTypesWithCommonGenericType((TypeBinding)superInterfaces[i], t);
        if (result != null)
          return result; 
      }  
    return null;
  }
  
  public TypeBinding getEquivalentOuterVariable(InferenceVariable variable, InferenceVariable[] outerVariables) {
    ThreeSets three = this.boundsPerVariable.get(variable);
    if (three != null)
      for (TypeBound bound : three.sameBounds) {
        byte b1;
        int j;
        InferenceVariable[] arrayOfInferenceVariable1;
        for (j = (arrayOfInferenceVariable1 = outerVariables).length, b1 = 0; b1 < j; ) {
          InferenceVariable iv = arrayOfInferenceVariable1[b1];
          if (TypeBinding.equalsEquals(bound.right, (TypeBinding)iv))
            return (TypeBinding)iv; 
          b1++;
        } 
      }  
    byte b;
    int i;
    InferenceVariable[] arrayOfInferenceVariable;
    for (i = (arrayOfInferenceVariable = outerVariables).length, b = 0; b < i; ) {
      InferenceVariable iv = arrayOfInferenceVariable[b];
      three = this.boundsPerVariable.get(outerVariables);
      if (three != null)
        for (TypeBound bound : three.sameBounds) {
          if (TypeBinding.equalsEquals(bound.right, (TypeBinding)variable))
            return (TypeBinding)iv; 
        }  
      b++;
    } 
    return null;
  }
}
